import graph

class score:
    """
    Abstract class implementing a scoring function.

    To obtain a working implementation, one has to define methods graph_score and data_score
    """
    
    sloops = True # do we allow self-loops in the networks ?
    delta = 10**(-7)
    data_factor = 1.0
    
    def __init__(self,n_disc):
        """
        n_disc - number of discretization levels for that gene
        """
        self.n_disc=n_disc
        
    def graph_score(self,n_vert,n_par,n_data):
        """
        Method for computing graph score factor - ABSTRACT method

        it computes the score given :
        n_vert - number of vertices in the network
        n_par - number of parents of a given gene
        n_data - number of data points
        """
        pass
    def data_min(self,v,data):
        """
        method for computing the lower bound of the data score factor

        parameters:
        v - number of the vertice
        data - list of datapoints :

        each datapoint is a vector containing 2*n_vert values (expression levels for all genes for time points t and t+1
        """
        return 0.0
    
    def data_score(self,v,par,data):
        """
        abstract method for computing the data score factor

        parameters:
        v - number of the vertice
        par - list of numbers of parents
        data - list of datapoints :

        each datapoint is a vector containing 2*n_vert values (expression levels for all genes for time points t and t+1
        """
        
        pass

    def sel_data(self,v,par,data,static=False):
        """
        given a vertice, list of his parents and data,
        select all lists of parents values at time t
        and child values at point t+1
        """
        n = len(data[0])/2
        r = []
        for i in data:
            r.append([])
        for i in par:
            for j,d in enumerate(data):
                r[j].append(d[i])
        for j,d in enumerate(data):
            r[j].append(d[v+n])
        r.sort()
        #print r
        return r

    def subsets(self,list,k):
        """
        generate all the k-long subsets of a list 
        """

        s = graph.stack()
        s.put((list,k,[]))
        try:
            while True: # exception used to get out of the loop
                list,k,acc = s.get()
                if k==0:
                    yield acc
                elif list==[]:
                    pass
                else:
                    s.put((list[1:],k,acc))
                    s.put((list[1:],k-1,acc+[list[0]]))
        except IndexError:
            pass # we've exhausted all options

    def stats(self,sd,v,par):
        def strec(key,prob,d):
            first,first_disc=d[0]
            rest=d[1:]
            if rest==[]:
                try:
                    stats_par[key]+=prob
                except KeyError:
                    stats_par[key]=prob
                if first_disc:
                    try:
                        stats_all[key+(first,)]+=prob
                    except KeyError:
                        stats_all[key+(first,)]=prob
                else:
                    try:
                        stats_all[key+(0,)]+=prob*(1-first)
                    except KeyError:
                        stats_all[key+(0,)]=prob*(1-first)
                    try:
                        stats_all[key+(1,)]+=prob*first
                    except KeyError:
                        stats_all[key+(1,)]=prob*first
            else:
                if first_disc:
                    strec(key+(first,),prob,rest)
                else:
                    strec(key+(0,),prob*(1-first),rest)
                    strec(key+(1,),prob*first,rest)
        d_disc=map(lambda p: p.n_disc,par+[v])
        if 0 not in d_disc:
            return self.stats_disc(sd)
        stats_all = {}
        stats_par = {}
        for d in sd:
            strec((),1,zip(d,d_disc))
        return stats_all,stats_par
        
    def stats_disc(self,sd):
        stats_all = {}
        stats_par = {}
        for d in sd:
            pars = tuple(d[:-1])
            all = tuple(d)
            try:
                stats_par[pars]+=1
            except KeyError:
                stats_par[pars]=1
            try:
                stats_all[all]+=1
            except KeyError:
                stats_all[all]=1
        return stats_all,stats_par
		    
    def learn_1(self,v,vert,data,n_points):
        if self.n_disc==None:
            return self.learn_alg1(v,vert,data,n_points)
        else:
            return self.learn_alg2(v,vert,data,n_points)
            
    def learn_alg1(self,v,vert,data,n_points):
        if not self.sloops and v in vert:
            import copy
            vv = copy.copy(vert)
            vv.remove(v)
            vert = vv
        n = len(vert)
        nd = n_points
        min = []
        mindata = self.data_min(v,data)
        mind = self.data_score(v,[],data)
        ming = self.graph_score_alg1(n,v.n_disc,[],nd)
        mins =  ming + mind
        min_set = [[]]
        #print mins,self.graph_score(n,size,nd)
        from pqueue import PQueue
        subsets=PQueue()
        for parent in vert:
            subsets.insert(self.graph_score_alg1(n,v.n_disc,[parent.n_disc],nd),[parent])
        while len(subsets):
            mg,sub=subsets.pop()
            if mins<=mindata+mg: #we cannot improve the score
                break
            md = self.data_score(v,sub,data)
            if mg + md < mins: #-self.delta:
                mins = mg+md
                ming = mg
                mind = md
                min = sub
                min_set = [sub]
            elif mg+md<mins+self.delta:
                min_set.append(sub)
            #insert sub's successors
            last_parent=vert.index(sub[-1])
            sub_succ=sub+[None]
            for parent in vert[last_parent+1:]:
                sub_succ[-1]=parent
                mg=self.graph_score_alg1(n,v.n_disc,map(lambda par:par.n_disc,sub_succ),nd)
                #if mins<mindata+mg: #we cannot improve the score
                #    break #with the remaining subsets, since parents in vert are sorted(!)
                subsets.insert(mg,sub_succ)
        #print min_set
        return min,mins

        
    
    def learn_alg2(self,v,vert,data,n_points):
        n = len(vert)
        nd = n_points
        min = []
        mindata = self.data_min(v,data)
        maxdata = self.data_score(v,[],data)
        mind = maxdata
        ming = self.graph_score(n,0,nd)
        mins =  ming + mind
        min_set = [[]]
        size = 1
        if not self.sloops and v in vert:
            import copy
            vv = copy.copy(vert)
            vv.remove(v)
            vert = vv
        #print mins,self.graph_score(n,size,nd)
        while mins-mindata>=self.graph_score(n,size,nd):
            #we can possibly improve the score
            mg = self.graph_score(n,size,nd)
            #print "subset size: %d, minscore: %g, graphscore: %g"%(size,mins,mg)

            for sub in self.subsets(vert,size):
                md = self.data_score(v,sub,data)
                #print "parents %s, data score %g, score: %g"%(sub,md,md+mg)
                if mg + md < mins: #-self.delta:
                    mins = mg+md
                    ming = mg
                    mind = md
                    min = sub
                    min_set = [sub]
                elif mg+md<mins+self.delta:
                    min_set.append(sub)
            size+=1
        #print min_set
        return min,mins

    def learn_all(self,vertices,data,n_points):
        par = {}
        for v in vertices:
            par[v] = self.learn_1(v,vertices,data,n_points)
        return par
    
    def score_graph(self,g,data):
        s = 0.0
        n_vert = len(g.vertices)
        for i,v in enumerate(g.vertices):
            p = g.parents(v)
            n,d = data.sel_data(i)
            #print v
            #print p
            #print len(d)
            sg = self.graph_score(n_vert,len(p),n)
            sd = self.data_score(i,p,d)
            s+=sg+sd
            #print sg,sd
        return s

class score_cont:
		    
    def stats(self,sd,v=None,par=[]):
        stats_all = {}
        stats_par = {}
        def strec(key,prob,d):
            first=d[0]
            rest=d[1:]
            if rest==[]:
                try:
                    stats_par[key]+=prob
                except KeyError:
                    stats_par[key]=prob
                try:
                    stats_all[key+(0,)]+=prob*(1-first)
                except KeyError:
                    stats_all[key+(0,)]=prob*(1-first)
                try:
                    stats_all[key+(1,)]+=prob*first
                except KeyError:
                    stats_all[key+(1,)]=prob*first
            else:
                strec(key+(0,),prob*(1-first),rest)
                strec(key+(1,),prob*first,rest)
                
        for d in sd:
            strec((),1,d)
        return stats_all,stats_par

